==> babelplot-2025.7.tar.gz: babelplot-2025.7/package/babelplot/type/backend.py <==
"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2022
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import importlib as mprt
import importlib.util as mput
import sys as s
import typing as h
from pathlib import Path as path_t

from _operator import itemgetter as ItemAt
from babelplot.constant.backend import BACKEND_FROM_ALIAS, DEFAULT_BACKEND
from babelplot.constant.path import BACKEND_CATALOG_PY_PATH
from babelplot.type.plot_type import plot_e, plot_function_h, plot_type_h
from babelplot.type.translator import arguments_translations_h
from logger_36 import L

# List, and not tuple, to allow for population by backends.
plot_functions_h = dict[plot_e, list[plot_function_h | None]]


@d.dataclass(slots=True, repr=False, eq=False)
class backend_t:
    path: path_t
    py_path: str
    plot_functions: plot_functions_h | None = None
    arguments_translations: arguments_translations_h | None = None

    def Activate(
        self,
        plot_functions: plot_functions_h,
        arguments_translations: arguments_translations_h,
        /,
    ) -> None:
        """"""
        self.plot_functions = plot_functions
        self.arguments_translations = arguments_translations

    def PlotFunction(
        self,
        type_: plot_e,
        frame_dim: int,
        /,
    ) -> plot_function_h:
        """"""
        if (output := self.plot_functions.get(type_)) is None:
            raise ValueError(f"Unknown plot {type_}.")

        if output.__len__() <= frame_dim - 1:
            raise ValueError(f"Unknown plot {type_} for dimension {frame_dim}.")

        output = output[frame_dim - 1]
        if output is None:
            raise ValueError(f"Unknown plot {type_} for dimension {frame_dim}.")

        return output

    def TranslatedArguments(
        self,
        args: h.Sequence[h.Any],
        kwargs: dict[str, h.Any],
        who_s_asking: str | h.Callable | None,
        /,
    ) -> tuple[h.Sequence[h.Any], dict[str, h.Any]]:
        """"""
        translations = self.arguments_translations
        if (translations is None) or (translations.__len__() == 0):
            return args, kwargs

        out_args = []
        out_kwargs = {}

        for idx, value in enumerate(args):
            if (translation := translations.get((who_s_asking, idx), -1)) == -1:
                # No translation needed; Keeping the passed arg.
                out_args.append(value)
            elif translation is None:
                # The argument should be discarded.
                pass
            else:  # isinstance(translation, str)
                # The passed arg should actually be a kwarg with name "translation".
                out_kwargs[translation] = value

        from_kwargs_to_args = []
        for key, value in kwargs.items():
            # Search for specific translation (who_s_asking, key) before key alone.
            if (translation := translations.get((who_s_asking, key), -1)) == -1:
                if (translation := translations.get(key, -1)) == -1:
                    # No translation needed; Keeping the passed arg.
                    out_kwargs[key] = value
                elif translation is None:
                    # The argument should be discarded.
                    pass
                else:  # isinstance(translation, str)
                    out_kwargs[translation] = value
            elif translation is None:
                # The argument should be discarded.
                pass
            elif isinstance(translation, str):
                out_kwargs[translation] = value
            else:  # isinstance(translation, int)
                from_kwargs_to_args.append((translation, value))

        for idx, value in sorted(from_kwargs_to_args, key=ItemAt(0)):
            out_args.insert(idx, value)

        if (ApplyGlobalTranslation := translations.get(None)) is not None:
            ApplyGlobalTranslation(out_kwargs, who_s_asking)

        return out_args, out_kwargs

    def __str__(self) -> str:
        """"""
        return f"{self.path}, {self.py_path}, {self.plot_functions is not None}"


@d.dataclass(slots=True, repr=False, eq=False)
class backends_t(dict[str, backend_t]):

    def __post_init__(self) -> None:
        """"""
        parent = path_t(__file__).parent.parent.parent
        catalog_path = parent / path_t(*BACKEND_CATALOG_PY_PATH.split("."))
        for node in catalog_path.glob("*"):
            if node.name.startswith("_"):
                continue

            if node.is_dir():
                main = node / "main.py"
                if main.is_file():
                    node = main
                else:
                    continue

            py_path = ".".join(node.relative_to(parent).parent.parts + (node.stem,))

            self[node.stem[:-1]] = backend_t(path=node, py_path=py_path)

    def Activate(
        self, backend: str | path_t | None, /
    ) -> tuple[str | None, type | None]:
        """
        backends: Importing runtime BACKENDS currently creates an import cycle.
        """
        module = None
        if backend is None:
            backend = DEFAULT_BACKEND
        else:
            # Check if str | path_t corresponds to a file.
            path = path_t(backend)
            if path.is_file():
                # TODO: Check the exceptions that are raised when not a valid Python
                #     module (not found in the documentation).
                backend_stem = path.stem
                spec = mput.spec_from_file_location(backend_stem, path)

                module = mput.module_from_spec(spec)

                spec.loader.exec_module(module)
                s.modules[backend_stem] = module
            elif isinstance(backend, path_t):
                L.error(
                    f"{backend}: Unusable backend. "
                    f"Path does not point to a regular file."
                )
                return None, None

        if module is None:
            if backend in self:
                py_path = self[backend].py_path
            elif backend in BACKEND_FROM_ALIAS:
                py_path = self[BACKEND_FROM_ALIAS[backend]].py_path
            else:
                py_path = backend
            try:
                module = mprt.import_module(py_path)
            except ModuleNotFoundError as exception:
                L.error(
                    f"{backend}: Unusable backend. "
                    f"Might be due to backend not being installed "
                    f"or implementation error in backend:\n{exception}"
                )
                return None, None

        name = module.NAME
        plot_functions = module.PLOTS
        arguments_translations = getattr(module, "TRANSLATIONS", None)

        self[name].Activate(plot_functions, arguments_translations)

        return name, module.figure_t

    def PlotFunction(
        self,
        name: str,
        type_: plot_type_h | plot_function_h,
        frame_dim: int,
        /,
    ) -> h.Callable:
        """
        Returns the plot type callable for the given plot_e member and the dimension passed
        as "frame_dim". The available callables are passed as "plot_functions". The name of
        the backend, passed as "backend", is only used in error messages.
        """
        if isinstance(type_, str):
            type_ = plot_e.NewFromName(type_)

        if isinstance(type_, plot_e):
            return self[name].PlotFunction(type_, frame_dim)

        return type_

    def TranslatedArguments(
        self,
        name: str,
        args: h.Sequence[h.Any],
        kwargs: dict[str, h.Any],
        who_s_asking: str | h.Callable | None,
        /,
    ) -> tuple[h.Sequence[h.Any], dict[str, h.Any]]:
        """"""
        return self[name].TranslatedArguments(
            args,
            kwargs,
            who_s_asking,
        )

    def IsValid(self, name: str, /) -> bool:
        """"""
        return (name in self) or (name in BACKEND_FROM_ALIAS)

    def __str__(self) -> str:
        """"""
        names = sorted(filter(lambda _: _ not in BACKEND_FROM_ALIAS, self.keys()))
        names = str(names)[1:-1].replace("'", "")
        aliases = (f"{_als} -> {_nme}" for _als, _nme in BACKEND_FROM_ALIAS.items())

        output = [
            "Backends:",
            f"    Names: {names}",
            f"    Aliases: {', '.join(sorted(aliases))}",
            "",
        ]
        output.extend(f"    {_key}: {_vle}" for _key, _vle in self.items())

        return "\n".join(output)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""

==> babelplot-2025.8.tar.gz: babelplot-2025.8/package/babelplot/type/backend.py <==
"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2022
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import importlib as mprt
import importlib.util as mput
import sys as s
import typing as h
from pathlib import Path as path_t

from _operator import itemgetter as ItemAt
from babelplot.constant.backend import BACKEND_FROM_ALIAS, DEFAULT_BACKEND
from babelplot.constant.path import BACKEND_CATALOG_PY_PATH
from babelplot.type.plot_type import plot_e, plot_function_h, plot_type_h
from babelplot.type.translator import arguments_translations_h
from logger_36 import L

# List, and not tuple, to allow for population by backends.
plot_functions_h = dict[plot_e, list[plot_function_h | None]]


@d.dataclass(slots=True, repr=False, eq=False)
class backend_t:
    path: path_t
    py_path: str
    plot_functions: plot_functions_h | None = None
    arguments_translations: arguments_translations_h | None = None

    def Activate(
        self,
        plot_functions: plot_functions_h,
        arguments_translations: arguments_translations_h,
        /,
    ) -> None:
        """"""
        self.plot_functions = plot_functions
        self.arguments_translations = arguments_translations

    def PlotFunction(
        self,
        type_: plot_e,
        frame_dim: int,
        /,
    ) -> plot_function_h:
        """"""
        if (output := self.plot_functions.get(type_)) is None:
            L.error(f"Unknown plot {type_}.")
            return _FakePlotFunction(type_, frame_dim)

        if output.__len__() <= frame_dim - 1:
            L.error(f"Unknown plot {type_} for dimension {frame_dim}.")
            return _FakePlotFunction(type_, frame_dim)

        output = output[frame_dim - 1]
        if output is None:
            L.error(f"Unknown plot {type_} for dimension {frame_dim}.")
            return _FakePlotFunction(type_, frame_dim)

        return output

    def TranslatedArguments(
        self,
        args: h.Sequence[h.Any],
        kwargs: dict[str, h.Any],
        who_s_asking: str | h.Callable | None,
        /,
    ) -> tuple[h.Sequence[h.Any], dict[str, h.Any]]:
        """"""
        translations = self.arguments_translations
        if (translations is None) or (translations.__len__() == 0):
            return args, kwargs

        out_args = []
        out_kwargs = {}

        # TODO: Add the possibility to transform the arguments by letting "translation"
        #     being either a translation command or a tuple of a translation command and
        #     a conversion function. Useful for color format adaptation.
        for idx, value in enumerate(args):
            if (translation := translations.get((who_s_asking, idx), -1)) == -1:
                # No translation needed; Keeping the passed arg.
                out_args.append(value)
            elif translation is None:
                # The argument should be discarded.
                pass
            else:  # isinstance(translation, str)
                # The passed arg should actually be a kwarg with name "translation".
                out_kwargs[translation] = value

        from_kwargs_to_args = []
        for key, value in kwargs.items():
            # Search for specific translation (who_s_asking, key) before key alone.
            if (translation := translations.get((who_s_asking, key), -1)) == -1:
                if (translation := translations.get(key, -1)) == -1:
                    # No translation needed; Keeping the passed arg.
                    out_kwargs[key] = value
                elif translation is None:
                    # The argument should be discarded.
                    pass
                else:  # isinstance(translation, str)
                    out_kwargs[translation] = value
            elif translation is None:
                # The argument should be discarded.
                pass
            elif isinstance(translation, str):
                out_kwargs[translation] = value
            else:  # isinstance(translation, int)
                from_kwargs_to_args.append((translation, value))

        for idx, value in sorted(from_kwargs_to_args, key=ItemAt(0)):
            out_args.insert(idx, value)

        if (ApplyGlobalTranslation := translations.get(None)) is not None:
            ApplyGlobalTranslation(out_kwargs, who_s_asking)

        return out_args, out_kwargs

    def __str__(self) -> str:
        """"""
        return f"{self.path}, {self.py_path}, {self.plot_functions is not None}"


@d.dataclass(slots=True, repr=False, eq=False)
class backends_t(dict[str, backend_t]):

    def __post_init__(self) -> None:
        """"""
        parent = path_t(__file__).parent.parent.parent
        catalog_path = parent / path_t(*BACKEND_CATALOG_PY_PATH.split("."))
        for node in catalog_path.glob("*"):
            if node.name.startswith("_"):
                continue

            if node.is_dir():
                main = node / "main.py"
                if main.is_file():
                    node = main
                else:
                    continue

            py_path = ".".join(node.relative_to(parent).parent.parts + (node.stem,))

            self[node.stem[:-1]] = backend_t(path=node, py_path=py_path)

    def Activate(
        self, backend: str | path_t | None, /
    ) -> tuple[str | None, type | None]:
        """
        backends: Importing runtime BACKENDS currently creates an import cycle.
        """
        module = None
        if backend is None:
            backend = DEFAULT_BACKEND
        else:
            # Check if str | path_t corresponds to a file.
            path = path_t(backend)
            if path.is_file():
                # TODO: Check the exceptions that are raised when not a valid Python
                #     module (not found in the documentation).
                backend_stem = path.stem
                spec = mput.spec_from_file_location(backend_stem, path)

                module = mput.module_from_spec(spec)

                spec.loader.exec_module(module)
                s.modules[backend_stem] = module
            elif isinstance(backend, path_t):
                L.error(
                    f"{backend}: Unusable backend. "
                    f"Path does not point to a regular file."
                )
                return None, None

        if module is None:
            if backend in self:
                py_path = self[backend].py_path
            elif backend in BACKEND_FROM_ALIAS:
                py_path = self[BACKEND_FROM_ALIAS[backend]].py_path
            else:
                py_path = backend
            try:
                module = mprt.import_module(py_path)
            except ModuleNotFoundError as exception:
                L.error(
                    f"{backend}: Unusable backend. "
                    f"Might be due to backend not being installed "
                    f"or implementation error in backend:\n{exception}"
                )
                return None, None

        name = module.NAME
        plot_functions = module.PLOTS
        arguments_translations = getattr(module, "TRANSLATIONS", None)

        self[name].Activate(plot_functions, arguments_translations)

        return name, module.figure_t

    def PlotFunction(
        self,
        name: str,
        type_: plot_type_h | plot_function_h,
        frame_dim: int,
        /,
    ) -> h.Callable:
        """
        Returns the plot type callable for the given plot_e member and the dimension passed
        as "frame_dim". The available callables are passed as "plot_functions". The name of
        the backend, passed as "backend", is only used in error messages.
        """
        if isinstance(type_, str):
            type_ = plot_e.NewFromName(type_)

        if isinstance(type_, plot_e):
            return self[name].PlotFunction(type_, frame_dim)

        return type_

    def TranslatedArguments(
        self,
        name: str,
        args: h.Sequence[h.Any],
        kwargs: dict[str, h.Any],
        who_s_asking: str | h.Callable | None,
        /,
    ) -> tuple[h.Sequence[h.Any], dict[str, h.Any]]:
        """"""
        return self[name].TranslatedArguments(
            args,
            kwargs,
            who_s_asking,
        )

    def IsValid(self, name: str, /) -> bool:
        """"""
        return (name in self) or (name in BACKEND_FROM_ALIAS)

    def __str__(self) -> str:
        """"""
        names = sorted(filter(lambda _: _ not in BACKEND_FROM_ALIAS, self.keys()))
        names = str(names)[1:-1].replace("'", "")
        aliases = (f"{_als} -> {_nme}" for _als, _nme in BACKEND_FROM_ALIAS.items())

        output = [
            "Backends:",
            f"    Names: {names}",
            f"    Aliases: {', '.join(sorted(aliases))}",
            "",
        ]
        output.extend(f"    {_key}: {_vle}" for _key, _vle in self.items())

        return "\n".join(output)


def _FakePlotFunction(type_: plot_e, frame_dim: int, /) -> h.Callable[..., None]:
    """"""

    def Actual(*args, **kwargs) -> None:
        #
        args = ", ".join(map(lambda _: type(_).__name__, args))
        kwargs = ", ".join(
            f"{_key}={type(_vle).__name__}" for _key, _vle in kwargs.items()
        )
        L.warning(f"Unhandled plot request: {type_.name}.{frame_dim}: {args}, {kwargs}")

    return Actual


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""
